@@ -1,7 +1,27 @@
 use v6;
 use fatal;
 
-unit class Getopt::Long:ver<0.1.5>;
+unit class Getopt::Long:ver<0.1.6>;
+
+role Exceptional is CORE::Exception {
+}
+
+class Exception does Exceptional {
+	has $.message;
+	method new(Str $message) {
+		return self.bless(:$message);
+	}
+}
+
+class ValueInvalid does Exceptional {
+	has $.format;
+	method new(Str $format) {
+		return self.bless(:$format);
+	}
+	method rethrow-with(Str $name) {
+		die Exception.new($!format.sprintf($name));
+	}
+}
 
 my sub null-converter(Str:D $value --> Str) {
 	return $value;
@@ -27,6 +47,29 @@ my sub complex-converter(Str:D $value --> Complex) {
 	return $value.Complex;
 }
 
+my sub io-converter(Str:D $value --> IO::Path) {
+	return $value.IO;
+	CATCH { default {
+		die ValueInvalid.new(qq{Can not convert %s argument "$value" a path: $_.message()});
+	}}
+}
+
+my sub datetime-converter(Str:D $value --> DateTime) {
+	return DateTime.new($value);
+
+	CATCH { when X::Temporal {
+		die ValueInvalid.new(qq{Can not convert %s argument "$value" to datetime: $_.message()});
+	}}
+}
+
+my sub date-converter(Str:D $value --> Date) {
+	return Date.new($value);
+
+	CATCH { when X::Temporal {
+		die ValueInvalid.new(qq{Can not convert %s argument "$value" to date: $_.message()});
+	}}
+}
+
 my sub maybe-converter(Str:D $value --> Any) {
 	return val($value);
 }
@@ -34,6 +77,10 @@ my sub maybe-converter(Str:D $value --> Any) {
 my role Store {
 	has Str:D $.key is required;
 	has Sub:D $.converter = &null-converter;
+	has Junction:D $.constraints = all();
+	method check-constraints($value) {
+		die ValueInvalid.new(qq{Can't accept %s argument "$value" because it fails its constraints}) unless $value ~~ $!constraints;
+	}
 	method store-convert($value, $hash) {
 		self.store-direct($!converter($value), $hash);
 	}
@@ -42,6 +89,7 @@ my role Store {
 
 my class ScalarStore does Store {
 	method store-direct(Any:D $value, Hash:D $hash) {
+		self.check-constraints($value);
 		$hash{$!key} = $value;
 	}
 }
@@ -55,6 +103,7 @@ my class CountStore does Store {
 my class ArrayStore does Store {
 	has Any:U $.type = $!converter.returns;
 	method store-direct(Any:D $value, Hash:D $hash) {
+		self.check-constraints($value);
 		$hash{$!key} //= $!type === Any ?? Array !! Array[$!type].new;
 		$hash{$!key}.push($value);
 	}
@@ -64,8 +113,10 @@ my class HashStore does Store {
 	has Any:U $.type = $!converter.returns;
 	method store-convert(Any:D $pair, Hash:D $hash) {
 		my ($key, $value) = $pair.split('=', 2);
+		my $converted-value = $!converter($value);
+		self.check-constraints($converted-value);
 		$hash{$!key} //= $!type === Any ?? Hash !! Hash[$!type].new;
-		$hash{$!key}{$key} = $!converter($value);
+		$hash{$!key}{$key} = $converted-value;
 	}
 	method store-direct(Any:D $pair, Hash:D $hash) {
 		!!!
@@ -85,9 +136,14 @@ my class Option {
 	}
 }
 
+has Code:D @!positionals;
 has Option:D %!options;
 
-submethod BUILD(:%!options) { }
+submethod BUILD(:@!positionals, :%!options) { }
+
+method !positionals {
+	return @!positionals.map(*.returns);
+}
 
 method !options {
 	return %!options.values;
@@ -99,9 +155,9 @@ my %store-for = (
 	''  => ScalarStore,
 );
 
-my sub make-option(@names, $multi-class, $multi-args, $arity, %options-args?, $negatable?) {
+my sub make-option(@names, Any:U $multi-class, %multi-args, Range $arity, %options-args?, Bool $negatable?) {
 	return flat @names.map: -> $name {
-		my $store = $multi-class.new(|%$multi-args, :key(@names[0]));
+		my $store = $multi-class.new(|%multi-args, :key(@names[0]));
 		my @options;
 		@options.push: Option.new(:$name, :$store, :$arity, :default, |%options-args);
 		if $negatable {
@@ -112,6 +168,20 @@ my sub make-option(@names, $multi-class, $multi-args, $arity, %options-args?, $n
 	}
 }
 
+my sub converter-for-format(Str:D $format) {
+	state %converter-for-format =
+		i => &int-converter,
+		o => &int-converter, # compatability with p5
+		s => &null-converter,
+		f => &num-converter,
+		r => &rat-converter,
+		c => &complex-converter,
+		p => &io-converter,
+		d => &datetime-converter,
+		a => &date-converter;
+	return %converter-for-format{$format} // die Exception.new("No such format $format");
+};
+
 my grammar Argument {
 	token TOP {
 		<names> <argument>
@@ -140,18 +210,9 @@ my grammar Argument {
 		{ make [ CountStore, {}, 0..0 ] }
 	}
 
-	my %converter-for-format = (
-		i => &int-converter,
-		o => &int-converter, # compatability with p5
-		s => &null-converter,
-		f => &num-converter,
-		r => &rat-converter,
-		c => &complex-converter,
-	);
-
 	token type {
-		<[siforc]>
-		{ make %converter-for-format{$/} }
+		<alpha>
+		{ make converter-for-format(~$/) }
 	}
 
 	token equals {
@@ -184,17 +245,7 @@ my grammar Argument {
 	}
 }
 
-role Exceptional {
-}
-
-class Exception is ::Exception does Exceptional {
-	has $.message;
-	method new(Str $message) {
-		return self.bless(:$message);
-	}
-}
-
-method new-from-patterns(@patterns, *%args) {
+method new-from-patterns(Getopt::Long:U: @patterns, Str :$positionals = "") {
 	my %options;
 	for @patterns -> $pattern {
 		if Argument.parse($pattern) -> $match {
@@ -206,18 +257,39 @@ method new-from-patterns(@patterns, *%args) {
 			die Exception.new("Couldn't parse '$pattern'");
 		}
 	}
-	return self.new(|%args, :%options);
+	my @positionals = $positionals.comb.map(&converter-for-format);
+	return self.new(:%options, :@positionals);
 }
 
-my %converter-for-type{Any:U} = (
-	(Int)     => &int-converter,
-	(Rat)     => &rat-converter,
-	(Num)     => &num-converter,
-	(Real)    => &real-converter,
-	(Complex) => &complex-converter,
-	(Str)     => &null-converter,
-	(Any)     => &maybe-converter,
-);
+sub get-converter(Any:U $type) {
+	state %converter-for-type{Any:U} = (
+		(Int)      => &int-converter,
+		(Rat)      => &rat-converter,
+		(Num)      => &num-converter,
+		(Real)     => &real-converter,
+		(Complex)  => &complex-converter,
+		(Str)      => &null-converter,
+		(IO::Path) => &io-converter,
+		(DateTime) => &datetime-converter,
+		(Date)     => &date-converter,
+		(Any)      => &maybe-converter,
+	);
+
+	if %converter-for-type{$type} -> &converter {
+		return &converter;
+	}
+	elsif $type.HOW ~~ Metamodel::EnumHOW {
+		my $valid-values = $type.WHO.keys.sort({ $type.WHO{$^value} }).join(", ");
+		sub enum-converter(Str $value) {
+			return $type.WHO{$value} // die ValueInvalid.new(qq{Can't convert %s argument "$value" to $type.^name(), valid values are: $valid-values});
+		}
+		trait_mod:<returns>(&enum-converter, $type);
+		return &enum-converter;
+	}
+	else {
+		die Exception.new("No conversion known for type {$type.^name}");
+	}
+}
 
 my role Formatted {
 	has Positional $.format is required;
@@ -232,30 +304,6 @@ multi sub trait_mod:<is>(Parameter $param, Str:D :$getopt!) is export(:DEFAULT,
 	}
 }
 
-my sub parse-parameter(Parameter $param) {
-	my @names = $param.named_names;
-	if $param ~~ Formatted {
-		return make-option(@names, |$param.format.list);
-	}
-	else {
-		if $param.sigil eq '$' {
-			my $type = $param.type;
-			if $param.type === Bool {
-				return make-option(@names, ScalarStore, {}, 0..0, {}, $param.default)
-			}
-			else {
-				my $converter = %converter-for-type{$param.type} // &null-converter;
-				return make-option(@names, ScalarStore, { :$converter }, 1..1);
-			}
-		}
-		else {
-			my $type = $param.type.of ~~ Any ?? $param.type.of !! Any;
-			my $converter = %converter-for-type{$type} // &null-converter;
-			return make-option(@names, %store-for{$param.sigil}, { :$type, :$converter }, 1..1);
-		}
-	}
-}
-
 my role Parsed {
 	has Getopt::Long:D $.getopt is required;
 }
@@ -264,27 +312,94 @@ multi sub trait_mod:<is>(Sub $sub, :$getopt!) is export(:DEFAULT, :traits) {
 	$sub does Parsed(Getopt::Long.new-from-sub($sub));
 }
 
-method new-from-sub(Sub $main) {
-	my multi get-options($candidate) {
-		return $candidate.signature.params.grep(*.named).map(&parse-parameter).flat;
+my multi get-positionals($candidate) {
+	return $candidate.signature.params.grep(*.positional).map(*.type);
+}
+
+my multi get-positionals(Parsed $candidate) {
+	return $candidate.getopt!positionals;
+}
+
+my multi get-named($candidate) {
+	my @options;
+	for $candidate.signature.params.grep(*.named) -> $param {
+		my @names = $param.named_names;
+		if $param ~~ Formatted {
+			@options.append: make-option(@names, |$param.format.list);
+		}
+		else {
+			if $param.sigil eq '$' {
+				my $type = $param.type;
+				my $constraints = $param.constraints;
+				if $param.type === Bool {
+					@options.append: make-option(@names, ScalarStore, { :$constraints }, 0..0, {}, ?$param.default)
+				}
+				else {
+					my $converter = get-converter($param.type);
+					@options.append: make-option(@names, ScalarStore, { :$converter, :$constraints }, 1..1);
+				}
+			}
+			else {
+				my $type = $param.type.of ~~ Any ?? $param.type.of !! Any;
+				my $converter = get-converter($type);
+				@options.append: make-option(@names, %store-for{$param.sigil}, { :$type, :$converter }, 1..1);
+			}
+		}
 	}
-	my multi get-options(Parsed $candidate) {
-		return $candidate.getopt!options;
-	}
-	my %options;
+	return @options;
+}
+my multi get-named(Parsed $candidate) {
+	return $candidate.getopt!options;
+}
+
+method new-from-sub(Getopt::Long:U: Sub $main) {
+	my (%options, @positional-types);
 	for $main.candidates -> $candidate {
-		for get-options($candidate) -> $option {
+		for get-named($candidate) -> $option {
 			if %options{$option.name}:exists and %options{$option.name} !eqv $option {
 				die Exception.new("Can't merge arguments for {$option.name}");
 			}
 			%options{$option.name} = $option;
 		}
+		@positional-types.push: get-positionals($candidate);
 	}
-	return self.new(:%options);
+	my $elem-max = max(@positional-types».elems);
+	my @positionals = (0 ..^ $elem-max).map: -> $index {
+		my @types = @positional-types.grep(* > $index)»[$index];
+		die Exception.new("Positional arguments are of different types {@types.perl}") unless [===] @types;
+		get-converter(@types[0]);
+	}
+	return self.new(:%options, :@positionals);
 }
 
-method get-options(@args is copy, :%hash, :named-anywhere(:$permute) = False, :$compat-builtin = False, :$bundling = !$compat-builtin, :$compat-singles = $compat-builtin, :$compat-negation = $compat-builtin, :$write-args) {
+method get-options(Getopt::Long:D: @args is copy, :%hash, :named-anywhere(:$permute) = False, :$compat-builtin = False, :$bundling = !$compat-builtin, :$compat-singles = $compat-builtin, :$compat-negation = $compat-builtin, :$compat-positional = $compat-builtin, :$write-args) {
 	my @list;
+
+	sub wrap-exceptions(Str $description, Str $value, &action) {
+		return action($value);
+		CATCH {
+			when ValueInvalid {
+				.rethrow-with($description);
+			}
+			when X::Str::Numeric {
+				$_ does role :: does Exceptional {
+					method message() {
+						qq{Cannot convert $description "$value" to number: $.reason};
+					}
+				}
+				.rethrow;
+			}
+			when X::Numeric::CannotConvert {
+				$_ does role :: does Exceptional {
+					method message() {
+						"Cannot convert $description $.source to {$.target // $.target.perl}: $.reason"
+					}
+				}
+				.rethrow;
+			}
+		}
+	}
+
 	while @args {
 		my $head = @args.shift;
 
@@ -295,27 +410,10 @@ method get-options(@args is copy, :%hash, :named-anywhere(:$permute) = False, :$
 		}
 
 		sub take-value($option, $value) {
-			$option.store($value, %hash);
-			$consumed++;
-
-			CATCH {
-				when X::Str::Numeric {
-					$_ does role :: does Exceptional {
-						method message() {
-							qq{Cannot convert --$option.name() argument "$value" to number: $.reason};
-						}
-					}
-					.rethrow;
-				}
-				when X::Numeric::CannotConvert {
-					$_ does role :: does Exceptional {
-						method message() {
-							"Cannot convert --{$option.name} argument $.source to {$.target // $.target.perl}: $.reason"
-						}
-					}
-					.rethrow;
-				}
-			}
+			wrap-exceptions("--$option.name()", $value, -> $value {
+				$option.store($value, %hash);
+				$consumed++;
+			});
 		}
 		sub take-args($option) {
 			while @args && $consumed < $option.arity.min {
@@ -386,7 +484,11 @@ method get-options(@args is copy, :%hash, :named-anywhere(:$permute) = False, :$
 		}
 	}
 	@$write-args = @list if $write-args;
-	return \(|@list.map(&val), |%hash);
+	state Str @labels = <first second third fourth fifth sixth seventh eighth nineth tenth>;
+	my &fallback-converter = $compat-positional ?? &maybe-converter !! &null-converter;
+	my @converters = |@!positionals, &fallback-converter, *;
+	@list = (@labels Z @list Z @converters).map: { wrap-exceptions(|@^args) }
+	return \(|@list, |%hash);
 }
 
 our sub get-options-from(@args, *@elements, :$overwrite, *%config) is export(:DEFAULT, :functions) {
@@ -532,6 +634,29 @@ following line in your program:
 This will load the core of the Getopt::Long module and prepare your
 program for using it.
 
+=head2 Getopt::Long as a MAIN wrapper
+
+Getopt::Long can be used as a argument parsing MAIN wrapper, replacing
+the builtin argument parsing. It will by default offer a Unix-typical
+command line interface, but various options allow it to be more similar
+to Raku's ideosyncratic parsing.
+
+It supports the following types for named and positional arguments:
+
+=item Bool
+=item Any
+=item Str
+=item Int
+=item Rat
+=item Num
+=item Real
+=item Complex
+=item IO::Path
+=item DateTime
+=item Date
+
+It also supports any enum type.
+
 =head2 Simple options
 
 The most simple options are the ones that take no values. Their mere
@@ -582,7 +707,7 @@ or:
 Now, using C<--verbose> on the command line will enable
 C<$verbose>, as expected. But it is also allowed to use
 C<--noverbose> or C<--no-verbose>, which will disable
-C<< $options<verbose> >> by setting its value to C<False>.
+C<< $verbose >> by setting its value to C<False>.
 
 An incremental option is specified with a plus C<+> after the
 option name:
@@ -776,9 +901,6 @@ it with "no" or "no-". E.g. C<"foo!"> will allow C<--foo> (a value of
 0 will be assigned). If the option has aliases, this applies to the
 aliases as well.
 
-Using negation on a single letter option when bundling is in effect is
-pointless and will result in a warning.
-
 =end item
 
 =begin item
@@ -802,7 +924,7 @@ are:
 =begin item2
 s
 
-String. An arbitrary sequence of characters. It is valid for the
+String(C<Str>). An arbitrary sequence of characters. It is valid for the
 argument to start with C<-> or C<-->.
 
 =end item2
@@ -810,7 +932,7 @@ argument to start with C<-> or C<-->.
 =begin item2
 i
 
-Integer. This can be either an optional leading plus or minus sign,
+Integer (C<Int>). This can be either an optional leading plus or minus sign,
 followed by a sequence of digits, or an octal string (C<0o>, optionally
 followed by '0', '1', .. '7'), or a hexadecimal string (C<0x> followed
 by '0' .. '9', 'a' .. 'f', case insensitive), or a binary string (C<0b>
@@ -821,21 +943,42 @@ followed by a series of '0' and '1').
 =begin item2
 r
 
-Rational number. For example C<3.14>.
+Rational number (C<Rat>). For example C<3.14>.
 
 =end item2
 
 =begin item2
 f
 
-Floating-pointer number. For example C<3.14>, C<-6.23E24> and so on.
+Floating-pointer number (C<Num>). For example C<3.14>, C<-6.23E24> and so on.
 
 =end item2
 
 =begin item2
 c
 
-Complex number. For example C<1+2i>.
+Complex number (C<Complex>). For example C<1+2i>.
+
+=end item2
+
+=begin item2
+p
+
+Path (C<IO::Path>). For example C<foo/bar.txt>.
+
+=end item2
+
+=begin item2
+d
+
+A date and time (C<DateTime>). For example C<2019-12-30T01:23:45-0700>.
+
+=end item2
+
+=begin item2
+a
+
+A Date (C<Date>). For example C<2019-12-30>.
 
 =end item2
 
@@ -912,33 +1055,12 @@ The following two calls behave identically:
 
 =head2 Bundling
 
-With bundling it is possible to set several single-character options
-at once. For example if C<a>, C<v> and C<x> are all valid options,
-
-    -vax
-
-will set all three.
-
-Getopt::Long supports three styles of bundling. To enable bundling, a
-call to Getopt::Long::Configure is required.
-
-Configured this way, single-character options can be bundled but long
-options B<must> always start with a double dash C<--> to avoid
-ambiguity. For example, when C<vax>, C<a>, C<v> and C<x> are all valid
-options,
-
-    -vax
-
-will set C<a>, C<v> and C<x>, but
-
-    --vax
-
-will set C<vax>.
 
 =head1 Configuring Getopt::Long
 
 C<get-options> and C<get-options-from> take the following named options
-to configure.
+to configure. When using Getopt::Long as a C<MAIN> wrapper, you can set
+them using the C<%*SUB-MAIN-OPTS> variable:
 
 =begin item
 permute (default:disabled)
@@ -961,7 +1083,7 @@ compat-builtin (default: disabled)
 
 Enable all compatibility options that make argument parsing more like
 the builtin argument parsing. Currently that means disabling C<bundling>
-and enabling C<compat-singles> and C<compat-negation>.
+and enabling C<compat-singles>, C<compat-negation> and C<compat-positional>.
 
 =end item
 
@@ -1003,12 +1125,17 @@ typically by using the C<!> modifier.
 
 =end item
 
+=begin item
+compat-positional (default: C<$compat-builtin>)
+
+Enabling this will turn all positional arguments into allomorphs, if
+possible.
+
+=end item
+
 =head1 Return values and Errors
 
 C<get-options> returns a capture to indicate success.
-Configuration errors and errors in the option definitions are
-signalled using die() and will terminate the calling program unless
-caught by exception handling.
 
 =head1 Troubleshooting
 
